/*
    NetWinder Floating Point Emulator
    (c) Corel Computer Corporation, 1998

    Direct questions, comments to Scott Bambrough <scottb@corelcomputer.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "config.h"
#include "milieu.h"
#include "softfloat.h"
#include "fpopcode.h"
#include "fpa11.h"
#include "fpa11.inl"
#include "fpmodule.h"
#include "fpmodule.inl"

flag floatx80_is_nan(floatx80);

void SetRoundingMode(const unsigned int opcode);

unsigned int PerformFLT(const unsigned int opcode);
unsigned int PerformFIX(const unsigned int opcode);

unsigned int
PerformComparison(const unsigned int opcode,const unsigned int bCode);

unsigned int
PerformComparisonOperation(floatx80 Fn,floatx80 Fm,const unsigned int bCode);

unsigned int EmulateCPRT(const unsigned int opcode)
{
  unsigned int nRc = 1;

  //fp_printk("EmulateCPRT(0x%08x)\n",opcode);
  
  switch (opcode & MASK_CPRT_CODE)
  {
    case  FLT_CODE: nRc = PerformFLT(opcode); break;
    case  FIX_CODE: nRc = PerformFIX(opcode); break;
    
    case  WFS_CODE: writeFPSR(readRegister(getRd(opcode))); break;
    case  RFS_CODE: writeRegister(getRd(opcode),readFPSR()); break;

    case  WFC_CODE:
    {
       int mode = 0;
       __asm__ volatile ("mrs %0, cpsr; and %0, %0, #0x1f;" : : "g" (mode));
       nRc = (0x13 == mode) ? 1 : 0;	/* in SVC processor mode? */
       if (nRc) writeFPCR(readRegister(getRd(opcode)));
    }
    break;
    
    case  RFC_CODE:
    {
       int mode = 0;
       __asm__ volatile ("mrs %0, cpsr; and %0, %0, #0x1f;" : : "g" (mode));
       nRc = (0x13 == mode) ? 1 : 0;	/* in SVC processor mode? */
       if (nRc) writeRegister(getRd(opcode),readFPCR()); break;
    }
    break;

    case  CMF_CODE: nRc = PerformComparison(opcode,CMF_CODE); break;
    case  CNF_CODE: nRc = PerformComparison(opcode,CNF_CODE); break;
    case CMFE_CODE: nRc = PerformComparison(opcode,CMFE_CODE); break;
    case CNFE_CODE: nRc = PerformComparison(opcode,CNFE_CODE); break;
    
           default: nRc = 0;
  }
  
  return nRc;
}

unsigned int PerformFLT(const unsigned int opcode)
{
   unsigned int nRc = 1;
   SetRoundingMode(opcode);
   SetRoundingPrecision(opcode);
   
   switch (opcode & MASK_ROUNDING_PRECISION)
   {
      case ROUND_SINGLE:
      {
        fpa11.fpreg[getFn(opcode)].fType = typeSingle;
        fpa11.fpreg[getFn(opcode)].fValue.fSingle =
	   int32_to_float32(readRegister(getRd(opcode)));
      }
      break;

      case ROUND_DOUBLE:
      {
        fpa11.fpreg[getFn(opcode)].fType = typeDouble;
        fpa11.fpreg[getFn(opcode)].fValue.fDouble =
            int32_to_float64(readRegister(getRd(opcode)));
      }
      break;
        
      case ROUND_EXTENDED:
      {
        fpa11.fpreg[getFn(opcode)].fType = typeExtended;
        fpa11.fpreg[getFn(opcode)].fValue.fExtended =
	   int32_to_floatx80(readRegister(getRd(opcode)));
      }
      break;
      
      default: nRc = 0;
  }
  
  return nRc;
}

unsigned int PerformFIX(const unsigned int opcode)
{
   unsigned int nRc = 1;
   unsigned int Fn = getFm(opcode);
   
   SetRoundingMode(opcode);

   switch (fpa11.fpreg[Fn].fType)
   {
      case typeSingle:
      {
         writeRegister(getRd(opcode),
	               float32_to_int32(fpa11.fpreg[Fn].fValue.fSingle));
      }
      break;

      case typeDouble:
      {
         writeRegister(getRd(opcode),
	               float64_to_int32(fpa11.fpreg[Fn].fValue.fDouble));
      }
      break;
      	               
      case typeExtended:
      {
         writeRegister(getRd(opcode),
	               floatx80_to_int32(fpa11.fpreg[Fn].fValue.fExtended));
      }
      break;
      
      default: nRc = 0;
  }
  
  return nRc;
}

/* This instruction sets the flags N, Z, C, V in the FPSR. */
   
unsigned int PerformComparison(const unsigned int opcode, 
			       const unsigned int bCode)
{
   unsigned int Fn, Fm;
   floatx80 rFn, rFm;

   //fp_printk("PerformComparison(0x%08x,0x%08x)\n",opcode,bCode);

   Fn = getFn(opcode);
   Fm = getFm(opcode);
   
   //fp_printk("Fn = r%d which contains a ",Fn);
   switch (fpa11.fpreg[Fn].fType)
   {
      case typeSingle: 
        //fp_printk("single.\n");
        rFn = float32_to_floatx80(fpa11.fpreg[Fn].fValue.fSingle);
      break;

      case typeDouble: 
        //fp_printk("double.\n");
        rFn = float64_to_floatx80(fpa11.fpreg[Fn].fValue.fDouble);
      break;
      
      case typeExtended: 
        //fp_printk("extended.\n");
        rFn = fpa11.fpreg[Fn].fValue.fExtended;
      break;
      
      default: return 0;
   }

   if (CONSTANT_FM(opcode))
   {
     //fp_printk("Fm is a constant: #%d.\n",Fm);
     rFm = getExtendedConstant(Fm);
   }
   else
   {
     //fp_printk("Fm = r%d which contains a ",Fm);
      switch (fpa11.fpreg[Fm].fType)
      {
         case typeSingle: 
           //fp_printk("single.\n");
           rFm = float32_to_floatx80(fpa11.fpreg[Fm].fValue.fSingle);
         break;

         case typeDouble: 
           //fp_printk("double.\n");
           rFm = float64_to_floatx80(fpa11.fpreg[Fm].fValue.fDouble);
         break;
      
         case typeExtended: 
           //fp_printk("extended.\n");
           rFm = fpa11.fpreg[Fm].fValue.fExtended;
         break;
      
         default: return 0;
      }
   }

   if ((CNF_CODE == bCode) || (CNFE_CODE == bCode))
   {
     //fp_printk("Negating Fm...\n");
     rFm.high ^= 0x8000;
   }

   return PerformComparisonOperation(rFn,rFm,bCode);
}
   
unsigned int
PerformComparisonOperation(floatx80 Fn, floatx80 Fm,
			   const unsigned int bCode)
{
   //unsigned int *p;
   unsigned int flags = 0;

   //fp_printk("PerformComparisonOperation");
#if 0
   switch (bCode)
   {
      case CMF_CODE: fp_printk("(CMF)\n"); break;
      case CNF_CODE: fp_printk("(CNF)\n"); break;
      case CMFE_CODE: fp_printk("(CMFE)\n"); break;
      case CNFE_CODE: fp_printk("(CNFE)\n"); break;
   }
   p = (unsigned int *)&Fn;
   fp_printk("rFn = 0x%08x 0x%08x 0x%08x\n",p[0],p[1],p[2]);
   p = (unsigned int *)&Fm;
   fp_printk("rFm = 0x%08x 0x%08x 0x%08x\n",p[0],p[1],p[2]);
#endif
      
   switch (bCode)
   {
     case CMF_CODE:
     case CNF_CODE:
     {
        /* test for unordered condition */
        if ((1 == floatx80_is_signaling_nan(Fn)) ||
            (1 == floatx80_is_signaling_nan(Fm)))
          {
             //fp_printk("Invalid exception on comparision.\n");
             flags |= CC_OVERFLOW;
             if (BIT_AC & readFPSR()) flags |= CC_CARRY;
             float_raise(float_flag_invalid);
          }
     }
     break;

     case CMFE_CODE:
     case CNFE_CODE:
     {
        /* test for unordered condition */
        if ((1 == floatx80_is_nan(Fn)) ||
            (1 == floatx80_is_nan(Fm)) ||
            (1 == floatx80_is_signaling_nan(Fn)) ||
            (1 == floatx80_is_signaling_nan(Fm)))
          {
             //fp_printk("Invalid exception on comparision.\n");
             flags |= CC_OVERFLOW;
             if (BIT_AC & readFPSR()) flags |= CC_CARRY;
             float_raise(float_flag_invalid);
          }
     }
     break;
   }

   if (0 == flags)
   {
      //fp_printk("Flags: ");
      /* test for less than condition */
      if (1 == floatx80_lt(Fn,Fm))
      {
	//fp_printk("CC_NEGATIVE ");
         flags |= CC_NEGATIVE;
      }

      /* test for equal condition */
      if (1 == floatx80_eq(Fn,Fm))
      {
	//fp_printk("CC_ZERO     ");
        flags |= CC_ZERO;
      }

      /* test for greater than or equal condition */
      if (1 == floatx80_lt(Fm,Fn))
      {
	//fp_printk("CC_CARRY    ");
        flags |= CC_CARRY;
      }
      //fp_printk(" (0x%08x)\n",flags);
   }
   
   writeConditionCodes(flags);
   return 1;
}

